########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import io
import unittest
import unittest.mock

import multidict

import aioxmpp.xso
import aioxmpp.xml

from aioxmpp.utils import namespaces

import aioxmpp.httpupload.xso as httpupload_xso


class TestNamespace(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(namespaces.xep0363_http_upload,
                         "urn:xmpp:http:upload:0")


class TestRequest(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            httpupload_xso.Request,
            aioxmpp.xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(httpupload_xso.Request.TAG,
                         (namespaces.xep0363_http_upload, "request"))

    def test_is_iq_payload(self):
        self.assertIn(
            httpupload_xso.Request.TAG,
            aioxmpp.IQ.CHILD_MAP,
        )

    def test_filename(self):
        self.assertIsInstance(
            httpupload_xso.Request.filename,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Request.filename.tag,
            (None, "filename"),
        )
        self.assertIsInstance(
            httpupload_xso.Request.filename.type_,
            aioxmpp.xso.String,
        )
        self.assertIs(httpupload_xso.Request.filename.default,
                      aioxmpp.xso.NO_DEFAULT)

    def test_size(self):
        self.assertIsInstance(
            httpupload_xso.Request.size,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Request.size.tag,
            (None, "size"),
        )
        self.assertIsInstance(
            httpupload_xso.Request.size.type_,
            aioxmpp.xso.Integer,
        )
        self.assertIs(httpupload_xso.Request.filename.default,
                      aioxmpp.xso.NO_DEFAULT)

    def test_content_type(self):
        self.assertIsInstance(
            httpupload_xso.Request.content_type,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Request.content_type.tag,
            (None, "content-type"),
        )
        self.assertIsInstance(
            httpupload_xso.Request.content_type.type_,
            aioxmpp.xso.String,
        )
        self.assertIs(httpupload_xso.Request.filename.default,
                      aioxmpp.xso.NO_DEFAULT)

    def test_init_requires_arguments(self):
        with self.assertRaisesRegex(TypeError, r"argument"):
            httpupload_xso.Request()

    def test_init(self):
        r = httpupload_xso.Request(
            "filename",
            1234,
            "content type",
        )

        self.assertEqual(r.filename, "filename")
        self.assertEqual(r.size, 1234)
        self.assertEqual(r.content_type, "content type")


class TestHeader(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            httpupload_xso.Header,
            aioxmpp.xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(httpupload_xso.Header.TAG,
                         (namespaces.xep0363_http_upload, "header"))

    def test_name(self):
        self.assertIsInstance(
            httpupload_xso.Header.name,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Header.name.tag,
            (None, "name"),
        )
        self.assertIsInstance(
            httpupload_xso.Header.name.type_,
            aioxmpp.xso.String,
        )
        self.assertIs(httpupload_xso.Header.name.default,
                      aioxmpp.xso.NO_DEFAULT)

    def test_value(self):
        self.assertIsInstance(
            httpupload_xso.Header.value,
            aioxmpp.xso.Text,
        )
        self.assertIsInstance(
            httpupload_xso.Header.name.type_,
            aioxmpp.xso.String,
        )


class TestHeaderType(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            httpupload_xso.HeaderType,
            aioxmpp.xso.AbstractElementType,
        ))

    def test_get_xso_types(self):
        self.assertCountEqual(
            httpupload_xso.HeaderType.get_xso_types(),
            [httpupload_xso.Header]
        )

    def test_unpack(self):
        t = httpupload_xso.HeaderType
        el = unittest.mock.Mock(spec=httpupload_xso.Header)
        el.name = unittest.mock.sentinel.name
        el.value = unittest.mock.sentinel.value
        self.assertEqual(
            t.unpack(el),
            (unittest.mock.sentinel.name, unittest.mock.sentinel.value)
        )

    def test_pack(self):
        t = httpupload_xso.HeaderType
        with unittest.mock.patch("aioxmpp.httpupload.xso.Header") as Header:
            result = t.pack((unittest.mock.sentinel.name,
                             unittest.mock.sentinel.value))

        Header.assert_called_once_with()
        self.assertEqual(result, Header())

        self.assertEqual(result.name, unittest.mock.sentinel.name)
        self.assertEqual(result.value, unittest.mock.sentinel.value)


class TestPut(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            httpupload_xso.Put,
            aioxmpp.xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(httpupload_xso.Put.TAG,
                         (namespaces.xep0363_http_upload, "put"))

    def test_url(self):
        self.assertIsInstance(
            httpupload_xso.Put.url,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Put.url.tag,
            (None, "url"),
        )
        self.assertIsInstance(
            httpupload_xso.Put.url.type_,
            aioxmpp.xso.String,
        )
        self.assertIs(httpupload_xso.Put.url.default, aioxmpp.xso.NO_DEFAULT)

    def test_headers(self):
        self.assertIsInstance(
            httpupload_xso.Put.headers,
            aioxmpp.xso.ChildValueMultiMap,
        )
        self.assertIs(
            httpupload_xso.Put.headers.type_,
            httpupload_xso.HeaderType
        )
        self.assertIs(
            httpupload_xso.Put.headers.mapping_type,
            multidict.MultiDict,
        )

    def test_parses_multiple_headers_correctly(self):
        data = (
            "<put xmlns='{}' url='foo'>"
            "<header name='Authorization'>v1</header>"
            "<header name='Expires'>v2</header>"
            "<header name='Authorization'>v3</header>"
            "</put>"
        ).format(namespaces.xep0363_http_upload)
        buf = io.BytesIO(data.encode("utf-8"))
        result = aioxmpp.xml.read_single_xso(buf, httpupload_xso.Put)

        self.assertCountEqual(
            result.headers.items(),
            [
                ("Authorization", "v1"),
                ("Expires", "v2"),
                ("Authorization", "v3"),
            ]
        )

    def test_xso_after_load_keeps_valid_headers_intact(self):
        headers = {
            "Authorization": "xyz",
            "Cookie": "xyz",
            "Expires": "xyz",
        }

        p = httpupload_xso.Put()
        p.headers.update(headers)
        p.xso_after_load()

        self.assertCountEqual(
            p.headers.items(),
            headers.items(),
        )

    def test_xso_after_load_strips_newlines(self):
        headers = {
            "Authorization": "abc\ndef",
        }

        p = httpupload_xso.Put()
        p.headers.update(headers)
        p.xso_after_load()

        self.assertCountEqual(
            p.headers.items(),
            {
                "Authorization": "abcdef",
            }.items()
        )

    def test_xso_after_load_removes_non_whitelisted_headers(self):
        headers = {
            "foo": "bar"
        }

        p = httpupload_xso.Put()
        p.headers.update(headers)
        p.xso_after_load()

        self.assertCountEqual(
            p.headers.items(),
            {}.items()
        )


class TestGet(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            httpupload_xso.Get,
            aioxmpp.xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(httpupload_xso.Get.TAG,
                         (namespaces.xep0363_http_upload, "get"))

    def test_url(self):
        self.assertIsInstance(
            httpupload_xso.Get.url,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            httpupload_xso.Get.url.tag,
            (None, "url"),
        )
        self.assertIsInstance(
            httpupload_xso.Get.url.type_,
            aioxmpp.xso.String,
        )
        self.assertIs(httpupload_xso.Get.url.default, aioxmpp.xso.NO_DEFAULT)


class TestSlot(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            httpupload_xso.Slot,
            aioxmpp.xso.XSO,
        ))

    def test_is_iq_payload(self):
        self.assertIn(
            httpupload_xso.Slot.TAG,
            aioxmpp.IQ.CHILD_MAP,
        )

    def test_tag(self):
        self.assertEqual(httpupload_xso.Slot.TAG,
                         (namespaces.xep0363_http_upload, "slot"))

    def test_put(self):
        self.assertIsInstance(
            httpupload_xso.Slot.put,
            aioxmpp.xso.Child,
        )
        self.assertCountEqual(
            httpupload_xso.Slot.put._classes,
            [httpupload_xso.Put]
        )

    def test_get(self):
        self.assertIsInstance(
            httpupload_xso.Slot.get,
            aioxmpp.xso.Child,
        )
        self.assertCountEqual(
            httpupload_xso.Slot.get._classes,
            [httpupload_xso.Get]
        )

    def test_validate_rejects_missing_put(self):
        s = httpupload_xso.Slot()
        s.get = unittest.mock.Mock(spec=httpupload_xso.Get)

        with self.assertRaisesRegex(ValueError, r"missing PUT information"):
            s.validate()

    def test_validate_rejects_missing_get(self):
        s = httpupload_xso.Slot()
        s.put = unittest.mock.Mock(spec=httpupload_xso.Put)

        with self.assertRaisesRegex(ValueError, r"missing GET information"):
            s.validate()

    def test_validate_passes_if_both_are_present(self):
        s = httpupload_xso.Slot()
        s.get = unittest.mock.Mock(spec=httpupload_xso.Get)
        s.put = unittest.mock.Mock(spec=httpupload_xso.Put)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.carbons.xso as carbons_xso
import aioxmpp.misc as misc_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0280_carbons_2,
            "urn:xmpp:carbons:2",
        )


class TestEnable(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(
            issubclass(carbons_xso.Enable, xso.XSO)
        )

    def test_tag(self):
        self.assertEqual(
            carbons_xso.Enable.TAG,
            (namespaces.xep0280_carbons_2, "enable"),
        )

    def test_is_iq_payload(self):
        self.assertIn(
            carbons_xso.Enable,
            aioxmpp.IQ.payload._classes,
        )


class TestDisable(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(
            issubclass(carbons_xso.Disable, xso.XSO)
        )

    def test_tag(self):
        self.assertEqual(
            carbons_xso.Disable.TAG,
            (namespaces.xep0280_carbons_2, "disable"),
        )

    def test_is_iq_payload(self):
        self.assertIn(
            carbons_xso.Disable,
            aioxmpp.IQ.payload._classes,
        )


class Test_CarbonsWrapper(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(
            issubclass(carbons_xso._CarbonsWrapper, xso.XSO)
        )

    def test_has_no_tag(self):
        self.assertFalse(hasattr(carbons_xso._CarbonsWrapper, "TAG"))

    def test_forwarded(self):
        self.assertIsInstance(
            carbons_xso._CarbonsWrapper.forwarded,
            xso.Child
        )

    def test_stanza_returns_None_if_forwarded_is_None(self):
        s = carbons_xso._CarbonsWrapper()
        self.assertIsNone(s.forwarded)
        self.assertIsNone(s.stanza)

    def test_stanza_returns_stanza_from_forwarded(self):
        s = carbons_xso._CarbonsWrapper()
        s.forwarded = misc_xso.Forwarded()
        s.forwarded.stanza = unittest.mock.sentinel.foo

        self.assertEqual(
            s.stanza,
            s.forwarded.stanza,
        )

    def test_setting_stanza_creates_forwarded(self):
        s = carbons_xso._CarbonsWrapper()
        self.assertIsNone(s.forwarded)

        s.stanza = unittest.mock.sentinel.foo

        self.assertIsInstance(
            s.forwarded,
            misc_xso.Forwarded,
        )

        self.assertEqual(
            s.forwarded.stanza,
            unittest.mock.sentinel.foo,
        )

    def test_setting_stanza_reuses_forwarded(self):
        forwarded = misc_xso.Forwarded()
        s = carbons_xso._CarbonsWrapper()
        s.forwarded = forwarded
        s.stanza = unittest.mock.sentinel.foo

        self.assertIs(
            s.forwarded,
            forwarded,
        )

        self.assertEqual(
            s.forwarded.stanza,
            unittest.mock.sentinel.foo,
        )


class TestSent(unittest.TestCase):
    def test_is_carbons_wrapper(self):
        self.assertTrue(
            issubclass(carbons_xso.Sent, carbons_xso._CarbonsWrapper)
        )

    def test_tag(self):
        self.assertEqual(
            carbons_xso.Sent.TAG,
            (namespaces.xep0280_carbons_2, "sent"),
        )


class TestReceived(unittest.TestCase):
    def test_is_carbons_wrapper(self):
        self.assertTrue(
            issubclass(carbons_xso.Received, carbons_xso._CarbonsWrapper)
        )

    def test_tag(self):
        self.assertEqual(
            carbons_xso.Received.TAG,
            (namespaces.xep0280_carbons_2, "received"),
        )


class TestMessage(unittest.TestCase):
    def test_xep0280_sent(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0280_sent,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.Message.xep0280_sent._classes,
            {
                carbons_xso.Sent,
            }
        )

    def test_xep0280_received(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0280_received,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.Message.xep0280_received._classes,
            {
                carbons_xso.Received,
            }
        )



########################################################################
# File name: structs.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.structs` --- Simple data holders for common data types
#####################################################################

These classes provide a way to hold structured data which is commonly
encountered in the XMPP realm.

Stanza types
============

.. currentmodule:: aioxmpp

.. autoclass:: IQType

.. autoclass:: MessageType

.. autoclass:: PresenceType

.. autoclass:: ErrorType

Jabber IDs
==========

.. autoclass:: JID(localpart, domain, resource)

Presence
========

.. autoclass:: PresenceShow

.. autoclass:: PresenceState

.. currentmodule:: aioxmpp.structs

Languages
=========

.. autoclass:: LanguageTag

.. autoclass:: LanguageRange

.. autoclass:: LanguageMap

Functions for working with language tags
----------------------------------------

.. autofunction:: basic_filter_languages

.. autofunction:: lookup_language

"""

import collections
import enum
import functools
import warnings

from .stringprep import nodeprep, resourceprep, nameprep


_USE_COMPAT_ENUM = True


class CompatibilityMixin:
    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if not _USE_COMPAT_ENUM:
            return super().__eq__(other)

        if super().__eq__(other) is True:
            return True
        if self.value == other:
            warnings.warn(
                "as of aioxmpp 1.0, {} members will not compare equal to their "
                "values".format(type(self).__name__),
                DeprecationWarning,
                stacklevel=2,
            )
            return True
        return False


class ErrorType(CompatibilityMixin, enum.Enum):
    """
    Enumeration for the :rfc:`6120` specified stanza error types.

    These error types reflect are actually more reflecting the error classes,
    but the attribute is called "type" nonetheless. For consistency, we are
    calling it "type" here, too.

    The following types are specified. The quotations in the member
    descriptions are from :rfc:`6120`, Section 8.3.2.

    .. attribute:: AUTH

       The ``"auth"`` error type:

          retry after providing credentials

       When converted to an exception, it uses :exc:`~.XMPPAuthError`.

    .. attribute:: CANCEL

       The ``"cancel"`` error type:

          do not retry (the error cannot be remedied)

       When converted to an exception, it uses :exc:`~.XMPPCancelError`.

    .. attribute:: CONTINUE

       The ``"continue"`` error type:

          proceed (the condition was only a warning)

       When converted to an exception, it uses
       :exc:`~.XMPPContinueError`.

    .. attribute:: MODIFY

       The ``"modify"`` error type:

          retry after changing the data sent

       When converted to an exception, it uses
       :exc:`~.XMPPModifyError`.

    .. attribute:: WAIT

       The ``"wait"`` error type:

          retry after waiting (the error is temporary)

       When converted to an exception, it uses (guess what)
       :exc:`~.XMPPWaitError`.

    :class:`ErrorType` members compare and hash equal to their values. For
    example::

      assert ErrorType.CANCEL == "cancel"
      assert "cancel" == ErrorType.CANCEL
      assert hash(ErrorType.CANCEL) == hash("cancel")

    .. deprecated:: 0.7

       This behaviour will cease with aioxmpp 1.0, and the first assertion will
       fail, the second may fail.

       Please see the Changelog for :ref:`api-changelog-0.7` for further details
       on how to upgrade your code efficiently.

    """

    AUTH = "auth"
    CANCEL = "cancel"
    CONTINUE = "continue"
    MODIFY = "modify"
    WAIT = "wait"


class MessageType(CompatibilityMixin, enum.Enum):
    """
    Enumeration for the :rfc:`6121` specified Message stanza types.

    .. seealso::

       :attr:`~.Message.type_`
          Type attribute of Message stanzas.


    Each member has the following meta-information:

    .. autoattribute:: is_error

    .. autoattribute:: is_request

    .. autoattribute:: is_response

    .. note::

       The :attr:`is_error`, :attr:`is_request` and :attr:`is_response`
       meta-information attributes share semantics across :class:`MessageType`,
       :class:`PresenceType` and :class:`IQType`. You are encouraged to exploit
       this in full duck-typing manner in generic stanza handling code.

    The following types are specified. The quotations in the member
    descriptions are from :rfc:`6121`, Section 5.2.2.

    .. attribute:: NORMAL

       The ``"normal"`` Message type:

          The message is a standalone message that is sent outside the context
          of a one-to-one conversation or groupchat, and to which it is
          expected that the recipient will reply.  Typically a receiving client
          will present a message of type "normal" in an interface that enables
          the recipient to reply, but without a conversation history.  The
          default value of the 'type' attribute is "normal".

       Think of it as somewhat similar to "E-Mail via XMPP".

    .. attribute:: CHAT

       The ``"chat"`` Message type:

          The message is sent in the context of a one-to-one chat session.
          Typically an interactive client will present a message of type "chat"
          in an interface that enables one-to-one chat between the two parties,
          including an appropriate conversation history.

    .. attribute:: GROUPCHAT

       The ``"groupchat"`` Message type:

          The message is sent in the context of a multi-user chat environment
          […].  Typically a receiving client will present a message of type
          "groupchat" in an interface that enables many-to-many chat between
          the parties, including a roster of parties in the chatroom and an
          appropriate conversation history.

    .. attribute:: HEADLINE

       The ``"headline"`` Message type:

          The message provides an alert, a notification, or other transient
          information to which no reply is expected (e.g., news headlines,
          sports updates, near-real-time market data, or syndicated content).
          Because no reply to the message is expected, typically a receiving
          client will present a message of type "headline" in an interface that
          appropriately differentiates the message from standalone messages,
          chat messages, and groupchat messages (e.g., by not providing the
          recipient with the ability to reply).

       Do not confuse this message type with the
       :attr:`~.Message.subject` member of Messages!

    .. attribute:: ERROR

       The ``"error"`` Message type:

          The message is generated by an entity that experiences an error when
          processing a message received from another entity […].  A client that
          receives a message of type "error" SHOULD present an appropriate
          interface informing the original sender regarding the nature of the
          error.

       This is the only message type which is used in direct response to
       another message, in the sense that the Stanza ID is preserved in the
       response.

    :class:`MessageType` members compare and hash equal to their values. For
    example::

      assert MessageType.CHAT == "chat"
      assert "chat" == MessageType.CHAT
      assert hash(MessageType.CHAT) == hash("chat")

    .. deprecated:: 0.7

       This behaviour will cease with aioxmpp 1.0, and the first assertion will
       fail, the second may fail.

       Please see the Changelog for :ref:`api-changelog-0.7` for further details
       on how to upgrade your code efficiently.

    """

    NORMAL = "normal"
    CHAT = "chat"
    GROUPCHAT = "groupchat"
    HEADLINE = "headline"
    ERROR = "error"

    @property
    def is_error(self):
        """
        True for the :attr:`ERROR` type, false for all others.
        """
        return self == MessageType.ERROR

    @property
    def is_response(self):
        """
        True for the :attr:`ERROR` type, false for all others.

        This is intended. Request/Response semantics do not really apply for
        messages, except that errors are generally in response to other
        messages.
        """
        return self == MessageType.ERROR

    @property
    def is_request(self):
        """
        False. See :attr:`is_response`.
        """
        return False


class PresenceType(CompatibilityMixin, enum.Enum):
    """
    Enumeration for the :rfc:`6121` specified Presence stanza types.

    .. seealso::

       :attr:`~.Presence.type_`
          Type attribute of Presence stanzas.

    Each member has the following meta-information:

    .. autoattribute:: is_error

    .. autoattribute:: is_request

    .. autoattribute:: is_response

    .. autoattribute:: is_presence_state

    .. note::

       The :attr:`is_error`, :attr:`is_request` and :attr:`is_response`
       meta-information attributes share semantics across :class:`MessageType`,
       :class:`PresenceType` and :class:`IQType`. You are encouraged to exploit
       this in full duck-typing manner in generic stanza handling code.


    The following types are specified. The quotes in the member descriptions
    are from :rfc:`6121`, Section 4.7.1.

    .. attribute:: ERROR

       The ``"error"`` Presence type:

          An error has occurred regarding processing of a previously sent
          presence stanza; if the presence stanza is of type "error", it MUST
          include an <error/> child element […].

       This is the only presence stanza type which is used in direct response
       to another presence stanza, in the sense that the Stanza ID is preserved
       in the response.

       In addition, :attr:`ERROR` presence stanzas may be seen during presence
       broadcast if inter-server communication fails.

    .. attribute:: PROBE

       The ``"probe"`` Presence type:

          A request for an entity's current presence; SHOULD be generated only
          by a server on behalf of a user.

       This should not be seen in client code.

    .. attribute:: SUBSCRIBE

       The ``"subscribe"`` Presence type:

          The sender wishes to subscribe to the recipient's presence.

    .. attribute:: SUBSCRIBED

       The ``"subscribed"`` Presence type:

          The sender has allowed the recipient to receive their presence.

    .. attribute:: UNSUBSCRIBE

       The ``"unsubscribe"`` Presence type:

          The sender is unsubscribing from the receiver's presence.

    .. attribute:: UNSUBSCRIBED

       The ``"unsubscribed"`` Presence type:

          The subscription request has been denied or a previously granted
          subscription has been canceled.

    .. attribute:: AVAILABLE

       The Presence type signalled with an absent type attribute:

          The absence of a 'type' attribute signals that the relevant entity is
          available for communication […].

    .. attribute:: UNAVAILABLE

       The ``"unavailable"`` Presence type:

          The sender is no longer available for communication.

    :class:`PresenceType` members compare and hash equal to their values. For
    example::

      assert PresenceType.PROBE == "probe"
      assert "probe" == PresenceType.PROBE
      assert hash(PresenceType.PROBE) == hash("probe")

    .. deprecated:: 0.7

       This behaviour will cease with aioxmpp 1.0, and the first assertion will
       fail, the second may fail.

       Please see the Changelog for :ref:`api-changelog-0.7` for further details
       on how to upgrade your code efficiently.
    """

    ERROR = "error"
    PROBE = "probe"
    SUBSCRIBE = "subscribe"
    SUBSCRIBED = "subscribed"
    UNAVAILABLE = "unavailable"
    UNSUBSCRIBE = "unsubscribe"
    UNSUBSCRIBED = "unsubscribed"
    AVAILABLE = None

    @property
    def is_error(self):
        """
        True for the :attr:`ERROR` type, false otherwise.
        """
        return self == PresenceType.ERROR

    @property
    def is_response(self):
        """
        True for the :attr:`ERROR` type, false otherwise.

        This is intended. Request/Response semantics do not really apply for
        presence stanzas, except that errors are generally in response to other
        presence stanzas.
        """
        return self == PresenceType.ERROR

    @property
    def is_request(self):
        """
        False. See :attr:`is_response`.
        """
        return False

    @property
    def is_presence_state(self):
        """
        True for the :attr:`AVAILABLE` and :attr:`UNAVAILABLE` types, false
        otherwise.

        Useful to discern presence state notifications from meta-stanzas
        regarding presence broadcast control.
        """
        return (self == PresenceType.AVAILABLE or
                self == PresenceType.UNAVAILABLE)


class IQType(CompatibilityMixin, enum.Enum):
    """
    Enumeration for the :rfc:`6120` specified IQ stanza types.

    .. seealso::

       :attr:`~.IQ.type_`
          Type attribute of IQ stanzas.

    Each member has the following meta-information:

    .. autoattribute:: is_error

    .. autoattribute:: is_request

    .. autoattribute:: is_response

    .. note::

       The :attr:`is_error`, :attr:`is_request` and :attr:`is_response`
       meta-information attributes share semantics across :class:`MessageType`,
       :class:`PresenceType` and :class:`IQType`. You are encouraged to exploit
       this in full duck-typing manner in generic stanza handling code.

    The following types are specified. The quotations in the member
    descriptions are from :rfc:`6120`, Section 8.2.3.

    .. attribute:: GET

       The ``"get"`` IQ type:

           The stanza requests information, inquires about what
           data is needed in order to complete further operations, etc.

       A :attr:`GET` IQ must contain a payload, via the
       :attr:`~.IQ.payload` attribute.

    .. attribute:: SET

       The ``"set"`` IQ type:

           The stanza provides data that is needed for an operation to be
           completed, sets new values, replaces existing values, etc.

       A :attr:`SET` IQ must contain a payload, via the
       :attr:`~.IQ.payload` attribute.

    .. attribute:: ERROR

       The ``"error"`` IQ type:

           The stanza reports an error that has occurred regarding processing
           or delivery of a get or set request[…].

       :class:`~.IQ` objects carrying the :attr:`ERROR` type usually
       have the :attr:`~.IQ.error` set to a :class:`~.stanza.Error`
       instance describing the details of the error.

       The :attr:`~.IQ.payload` attribute may also be set if the sender
       of the :attr:`ERROR` was kind enough to include the data which caused
       the problem.

    .. attribute:: RESULT

       The ``"result"`` IQ type:

           The stanza is a response to a successful get or set request.

       A :attr:`RESULT` IQ may contain a payload with more data.

    :class:`IQType` members compare and hash equal to their values. For
    example::

      assert IQType.GET == "get"
      assert "get" == IQType.GET
      assert hash(IQType.GET) == hash("get")

    .. deprecated:: 0.7

       This behaviour will cease with aioxmpp 1.0, and the first assertion will
       fail, the second may fail.

       Please see the Changelog for :ref:`api-changelog-0.7` for further details
       on how to upgrade your code efficiently.
    """

    GET = "get"
    SET = "set"
    ERROR = "error"
    RESULT = "result"

    @property
    def is_error(self):
        """
        True for the :attr:`ERROR` type, false otherwise.
        """
        return self == IQType.ERROR

    @property
    def is_request(self):
        """
        True for request types (:attr:`GET` and :attr:`SET`), false otherwise.
        """
        return self == IQType.GET or self == IQType.SET

    @property
    def is_response(self):
        """
        True for the response types (:attr:`RESULT` and :attr:`ERROR`), false
        otherwise.
        """
        return self == IQType.RESULT or self == IQType.ERROR


class JID(collections.namedtuple("JID", ["localpart", "domain", "resource"])):
    """
    A Jabber ID (JID). To construct a JID, either use the actual constructor,
    or use the :meth:`fromstr` class method.

    If `strict` is false, unassigned codepoints are allowed in any of the parts
    of the JID. In the future, other deviations from the respective stringprep
    profiles may be allowed, too.

    The idea is to use non-`strict` when output is received from outside and
    when it is reflected, following the old principle "be conservative in what
    you send and liberal in what you receive". Otherwise, strict checking
    should be enabled. This brings maximum interoperability.

    .. automethod:: fromstr

    Information about a JID:

    .. attribute:: localpart

       The localpart, stringprep’d from the argument to the constructor.

    .. attribute:: domain

       The domain, stringprep’d from the argument to the constructor.

    .. attribute:: resource

       The resource, stringprep’d from the argument to the constructor.

    .. autoattribute:: is_bare

    .. autoattribute:: is_domain

    :class:`JID` objects are immutable. To obtain a JID object with a changed
    property, use one of the following methods:

    .. automethod:: bare

    .. automethod:: replace(*, [localpart], [domain], [resource])
    """

    __slots__ = []

    def __new__(cls, localpart, domain, resource, *, strict=True):
        if localpart:
            localpart = nodeprep(
                localpart,
                allow_unassigned=not strict
            )
        if domain is not None:
            domain = nameprep(
                domain,
                allow_unassigned=not strict
            )
        if resource:
            resource = resourceprep(
                resource,
                allow_unassigned=not strict
            )

        if not domain:
            raise ValueError("domain must not be empty or None")
        if len(domain.encode("utf-8")) > 1023:
            raise ValueError("domain too long")
        if localpart is not None:
            if not localpart:
                raise ValueError("localpart must not be empty")
            if len(localpart.encode("utf-8")) > 1023:
                raise ValueError("localpart too long")
        if resource is not None:
            if not resource:
                raise ValueError("resource must not be empty")
            if len(resource.encode("utf-8")) > 1023:
                raise ValueError("resource too long")

        return super().__new__(cls, localpart, domain, resource)

    def replace(self, **kwargs):
        """
        Construct a new :class:`JID` object, using the values of the current
        JID. Use the arguments to override specific attributes on the new
        object.
        """

        new_kwargs = {}

        strict = kwargs.pop("strict", True)

        try:
            localpart = kwargs.pop("localpart")
        except KeyError:
            pass
        else:
            if localpart:
                localpart = nodeprep(
                    localpart,
                    allow_unassigned=not strict
                )
            new_kwargs["localpart"] = localpart

        try:
            domain = kwargs.pop("domain")
        except KeyError:
            pass
        else:
            if not domain:
                raise ValueError("domain must not be empty or None")
            new_kwargs["domain"] = nameprep(
                domain,
                allow_unassigned=not strict
            )

        try:
            resource = kwargs.pop("resource")
        except KeyError:
            pass
        else:
            if resource:
                resource = resourceprep(
                    resource,
                    allow_unassigned=not strict
                )
            new_kwargs["resource"] = resource

        if kwargs:
            raise TypeError("replace() got an unexpected keyword argument"
                            " {!r}".format(
                                next(iter(kwargs))))

        return super()._replace(**new_kwargs)

    def __str__(self):
        result = self.domain
        if self.localpart:
            result = self.localpart + "@" + result
        if self.resource:
            result += "/" + self.resource
        return result

    def bare(self):
        """
        Return the bare version of this JID as new :class:`JID` object.
        """
        return self.replace(resource=None)

    @property
    def is_bare(self):
        """
        :data:`True` if the JID is bare, i.e. has an empty :attr:`resource`
        part.
        """
        return not self.resource

    @property
    def is_domain(self):
        """
        :data:`True` if the JID is a domain, i.e. if both the :attr:`localpart`
        and the :attr:`resource` are empty.
        """
        return not self.resource and not self.localpart

    @classmethod
    def fromstr(cls, s, *, strict=True):
        """
        Obtain a :class:`JID` object by parsing a JID from the given string
        `s`.
        """
        nodedomain, sep, resource = s.partition("/")
        if not sep:
            resource = None

        localpart, sep, domain = nodedomain.partition("@")
        if not sep:
            domain = localpart
            localpart = None
        return cls(localpart, domain, resource, strict=strict)


@functools.total_ordering
class PresenceShow(CompatibilityMixin, enum.Enum):
    """
    Enumeration to support the ``show`` element of presence stanzas.

    The enumeration members support total ordering. The order is defined by
    relevance and is the following (from lesser to greater): :attr:`XA`,
    :attr:`AWAY`, :attr:`NONE`, :attr:`CHAT`, :attr:`DND`. The order is
    intended to be used to extract the most relevant resource e.g. in a roster.

    .. versionadded:: 0.8

    .. attribute:: XA
       :annotation: = "xa"

       .. epigraph::

          The entity or resource is away for an extended period (xa = "eXtended
          Away").

          -- :rfc:`6121`, Section 4.7.2.1

    .. attribute:: EXTENDED_AWAY
       :annotation: = "xa"

       Alias to :attr:`XA`.

    .. attribute:: AWAY
       :annotation: = "away"

       .. epigraph::

          The entity or resource is temporarily away.

          -- :rfc:`6121`, Section 4.7.2.1

    .. attribute:: NONE
       :annotation: = None

       Signifies absence of the ``show`` element.

    .. attribute:: PLAIN
       :annotation: = None

       Alias to :attr:`NONE`.

    .. attribute:: CHAT
       :annotation: = "chat"

       .. epigraph::

          The entity or resource is actively interested in chatting.

          -- :rfc:`6121`, Section 4.7.2.1

    .. attribute:: FREE_FOR_CHAT
       :annotation: = "chat"

       Alias to :attr:`CHAT`.

    .. attribute:: DND
       :annotation: = "dnd"

       .. epigraph::

          The entity or resource is busy (dnd = "Do Not Disturb").

          -- :rfc:`6121`, Section 4.7.2.1

    .. attribute:: DO_NOT_DISTURB
       :annotation: = "dnd"

       Alias to :attr:`DND`.

    """
    XA = "xa"
    EXTENDED_AWAY = "xa"
    AWAY = "away"
    NONE = None
    PLAIN = None
    CHAT = "chat"
    FREE_FOR_CHAT = "chat"
    DND = "dnd"
    DO_NOT_DISTURB = "dnd"

    def __lt__(self, other):
        try:
            w1 = self._WEIGHTS[self]
            w2 = self._WEIGHTS[other]
        except KeyError:
            return NotImplemented
        return w1 < w2


PresenceShow._WEIGHTS = {
    PresenceShow.XA: -2,
    PresenceShow.AWAY: -1,
    PresenceShow.NONE: 0,
    PresenceShow.CHAT: 1,
    PresenceShow.DND: 2,
}


@functools.total_ordering
class PresenceState:
    """
    Hold a presence state of an XMPP resource, as defined by the presence
    stanza semantics.

    `available` must be a boolean value, which defines whether the resource is
    available or not. If the resource is available, `show` may be set to one of
    ``"dnd"``, ``"xa"``, ``"away"``, :data:`None`, ``"chat"`` (it is a
    :class:`ValueError` to attempt to set `show` to a non-:data:`None` value if
    `available` is false).

    :class:`PresenceState` objects are ordered by their availability and by
    their show values. Non-availability sorts lower than availability, and for
    available presence states the order is in the order of valid values given
    for the `show` above.

    .. attribute:: available

       As per the argument to the constructor, converted to a :class:`bool`.

    .. attribute:: show

       As per the argument to the constructor.

    .. automethod:: apply_to_stanza

    .. automethod:: from_stanza

    :class:`PresenceState` objects are immutable.

    """

    __slots__ = ["_available", "_show"]

    def __init__(self, available=False, show=PresenceShow.NONE):
        super().__init__()
        if not available and show != PresenceShow.NONE:
            raise ValueError("Unavailable state cannot have show value")
        if not isinstance(show, PresenceShow):
            try:
                show = PresenceShow(show)
            except ValueError:
                raise ValueError("Not a valid show value") from None
            else:
                warnings.warn(
                    "as of aioxmpp 1.0, the show argument must use "
                    "PresenceShow instead of str",
                    DeprecationWarning,
                    stacklevel=2
                )

        self._available = bool(available)
        self._show = show

    @property
    def available(self):
        return self._available

    @property
    def show(self):
        return self._show

    def __lt__(self, other):
        my_key = (self.available, self.show)
        try:
            other_key = (other.available, other.show)
        except AttributeError:
            return NotImplemented
        return my_key < other_key

    def __eq__(self, other):
        try:
            return (self.available == other.available and
                    self.show == other.show)
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        more = ""
        if self.available:
            if self.show != PresenceShow.NONE:
                more = " available show={!r}".format(self.show)
            else:
                more = " available"
        return "<PresenceState{}>".format(more)

    def apply_to_stanza(self, stanza_obj):
        """
        Apply the properties of this :class:`PresenceState` to a
        :class:`~aioxmpp.Presence` `stanza_obj`. The
        :attr:`~aioxmpp.Presence.type_` and
        :attr:`~aioxmpp.Presence.show` attributes of the object will be
        modified to fit the values in this object.
        """
        if self.available:
            stanza_obj.type_ = PresenceType.AVAILABLE
        else:
            stanza_obj.type_ = PresenceType.UNAVAILABLE
        stanza_obj.show = self.show

    @classmethod
    def from_stanza(cls, stanza_obj, strict=False):
        """
        Create and return a new :class:`PresenceState` object which inherits
        the presence state as advertised in the given
        :class:`~aioxmpp.Presence` stanza.

        If `strict` is :data:`True`, the value of `show` is strictly checked,
        that is, it is required to be :data:`None` if the stanza indicates an
        unavailable state.

        The default is not to check this.
        """

        if not stanza_obj.type_.is_presence_state:
            raise ValueError("presence state stanza required")
        available = stanza_obj.type_ == PresenceType.AVAILABLE
        if not strict:
            show = stanza_obj.show if available else PresenceShow.NONE
        else:
            show = stanza_obj.show
        return cls(available=available, show=show)


@functools.total_ordering
class LanguageTag:
    """
    Implementation of a language tag. This may be a fully RFC5646 compliant
    implementation some day, but for now it is only very simplistic stub.

    There is no input validation of any kind.

    :class:`LanguageTag` instances compare and hash case-insensitively.

    .. automethod:: fromstr

    .. autoattribute:: match_str

    .. autoattribute:: print_str

    """

    __slots__ = ("_tag",)

    def __init__(self, *, tag=None):
        if not tag:
            raise ValueError("tag cannot be empty")

        self._tag = tag

    @property
    def match_str(self):
        """
        The string which is used for matching two lanugage tags. This is the
        lower-cased version of the :attr:`print_str`.
        """
        return self._tag.lower()

    @property
    def print_str(self):
        """
        The stringified language tag.
        """
        return self._tag

    @classmethod
    def fromstr(cls, s):
        """
        Create a language tag from the given string `s`.

        .. note::

           This is a stub implementation which merely refers to the given
           string as the :attr:`print_str` and derives the :attr:`match_str`
           from that.

        """
        return cls(tag=s)

    def __str__(self):
        return self.print_str

    def __eq__(self, other):
        try:
            return self.match_str == other.match_str
        except AttributeError:
            return False

    def __lt__(self, other):
        try:
            return self.match_str < other.match_str
        except AttributeError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self.match_str <= other.match_str
        except AttributeError:
            return NotImplemented

    def __hash__(self):
        return hash(self.match_str)

    def __repr__(self):
        return "<{}.{}.fromstr({!r})>".format(
            type(self).__module__,
            type(self).__qualname__,
            str(self))


class LanguageRange:
    """
    Implementation of a language range. This may be a fully RFC4647 compliant
    implementation some day, but for now it is only very simplistic stub.

    There is no input validation of any kind.

    :class:`LanguageRange` instances compare and hash case-insensitively.

    .. automethod:: fromstr

    .. automethod:: strip_rightmost

    .. autoattribute:: match_str

    .. autoattribute:: print_str

    """

    __slots__ = ("_tag",)

    def __init__(self, *, tag=None):
        if not tag:
            raise ValueError("range cannot be empty")

        self._tag = tag

    @property
    def match_str(self):
        """
        The string which is used for matching two lanugage tags. This is the
        lower-cased version of the :attr:`print_str`.
        """
        return self._tag.lower()

    @property
    def print_str(self):
        """
        The stringified language tag.
        """
        return self._tag

    @classmethod
    def fromstr(cls, s):
        """
        Create a language tag from the given string `s`.

        .. note::

           This is a stub implementation which merely refers to the given
           string as the :attr:`print_str` and derives the :attr:`match_str`
           from that.

        """
        if s == "*":
            return cls.WILDCARD

        return cls(tag=s)

    def __str__(self):
        return self.print_str

    def __eq__(self, other):
        try:
            return self.match_str == other.match_str
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.match_str)

    def __repr__(self):
        return "<{}.{}.fromstr({!r})>".format(
            type(self).__module__,
            type(self).__qualname__,
            str(self))

    def strip_rightmost(self):
        """
        Strip the rightmost part of the language range. If the new rightmost
        part is a singleton or ``x`` (i.e. starts an extension or private use
        part), it is also stripped.

        Return the newly created :class:`LanguageRange`.
        """

        parts = self.print_str.split("-")
        parts.pop()
        if parts and len(parts[-1]) == 1:
            parts.pop()
        return type(self).fromstr("-".join(parts))

LanguageRange.WILDCARD = LanguageRange(tag="*")


def basic_filter_languages(languages, ranges):
    """
    Filter languages using the string-based basic filter algorithm described in
    RFC4647.

    `languages` must be a sequence of :class:`LanguageTag` instances which are
    to be filtered.

    `ranges` must be an iterable which represent the basic language ranges to
    filter with, in priority order. The language ranges must be given as
    :class:`LanguageRange` objects.

    Return an iterator of languages which matched any of the `ranges`. The
    sequence produced by the iterator is in match order and duplicate-free. The
    first range to match a language yields the language into the iterator, no
    other range can yield that language afterwards.
    """

    if LanguageRange.WILDCARD in ranges:
        yield from languages
        return

    found = set()

    for language_range in ranges:
        range_str = language_range.match_str
        for language in languages:
            if language in found:
                continue

            match_str = language.match_str
            if match_str == range_str:
                yield language
                found.add(language)
                continue

            if len(range_str) < len(match_str):
                if     (match_str[:len(range_str)] == range_str and
                        match_str[len(range_str)] == "-"):
                    yield language
                    found.add(language)
                    continue


def lookup_language(languages, ranges):
    """
    Look up a single language in the sequence `languages` using the lookup
    mechansim described in RFC4647. If no match is found, :data:`None` is
    returned. Otherwise, the first matching language is returned.

    `languages` must be a sequence of :class:`LanguageTag` objects, while
    `ranges` must be an iterable of :class:`LanguageRange` objects.
    """

    for language_range in ranges:
        while True:
            try:
                return next(iter(basic_filter_languages(
                    languages,
                    [language_range])))
            except StopIteration:
                pass

            try:
                language_range = language_range.strip_rightmost()
            except ValueError:
                break


class LanguageMap(dict):
    """
    A :class:`dict` subclass specialized for holding :class:`LanugageTag`
    instances as keys.

    In addition to the interface provided by :class:`dict`, instances of this
    class also have the following methods:

    .. automethod:: lookup

    .. automethod:: any
    """

    def lookup(self, language_ranges):
        """
        Perform an RFC4647 language range lookup on the keys in the
        dictionary. `language_ranges` must be a sequence of
        :class:`LanguageRange` instances.

        Return the entry in the dictionary with a key as produced by
        `lookup_language`. If `lookup_language` does not find a match and the
        mapping contains an entry with key :data:`None`, that entry is
        returned, otherwise :class:`KeyError` is raised.
        """
        keys = list(self.keys())
        try:
            keys.remove(None)
        except ValueError:
            pass
        keys.sort()
        key = lookup_language(keys, language_ranges)
        return self[key]

    def any(self):
        """
        Returns any element from the language map, preferring the :data:`None`
        key if it is available.

        Guarantees to always return the same element for a map with the same
        keys, even if the keys are iterated over in a different order.
        """
        if not self:
            raise ValueError("any() on empty map")

        try:
            return self[None]
        except KeyError:
            return self[min(self)]



########################################################################
# File name: test_rfc6120.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.rfc6120 as rfc6120
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs
import aioxmpp.nonza as nonza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestBindFeature(unittest.TestCase):
    def test_registered_at_StreamFeatures(self):
        self.assertIn(
            rfc6120.BindFeature.TAG,
            nonza.StreamFeatures.CHILD_MAP
        )


class TestBind(unittest.TestCase):
    def test_declare_ns(self):
        self.assertDictEqual(
            rfc6120.Bind.DECLARE_NS,
            {
                None: namespaces.rfc6120_bind,
            }
        )

    def test_tag(self):
        self.assertEqual(
            rfc6120.Bind.TAG,
            (namespaces.rfc6120_bind, "bind")
        )

    def test_jid(self):
        self.assertIsInstance(
            rfc6120.Bind.jid,
            xso.ChildText
        )
        self.assertEqual(
            rfc6120.Bind.jid.tag,
            (namespaces.rfc6120_bind, "jid")
        )
        self.assertIsInstance(
            rfc6120.Bind.jid.type_,
            xso.JID
        )
        self.assertIs(
            rfc6120.Bind.jid.default,
            None
        )

    def test_resource(self):
        self.assertIsInstance(
            rfc6120.Bind.resource,
            xso.ChildText
        )
        self.assertEqual(
            rfc6120.Bind.resource.tag,
            (namespaces.rfc6120_bind, "resource")
        )
        self.assertIs(
            rfc6120.Bind.resource.default,
            None
        )

    def test_default_init(self):
        obj = rfc6120.Bind()
        self.assertIsNone(obj.jid)
        self.assertIsNone(obj.resource)

    def test_init(self):
        jid = structs.JID.fromstr("foo@bar.example")
        obj = rfc6120.Bind(
            jid=jid,
            resource="foobar"
        )
        self.assertEqual(
            jid,
            obj.jid)
        self.assertEqual(
            "foobar",
            obj.resource)

    def test_registered_at_IQ(self):
        self.assertIn(
            rfc6120.Bind.TAG,
            stanza.IQ.CHILD_MAP
        )



########################################################################
# File name: markers.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso

from ..stanza import Message

from aioxmpp.utils import namespaces

namespaces.xep0333_markers = "urn:xmpp:chat-markers:0"


class Marker(aioxmpp.xso.XSO):
    id_ = aioxmpp.xso.Attr(
        "id"
    )


class ReceivedMarker(Marker):
    TAG = (namespaces.xep0333_markers, "received")


class DisplayedMarker(Marker):
    TAG = (namespaces.xep0333_markers, "displayed")


class AcknowledgedMarker(Marker):
    TAG = (namespaces.xep0333_markers, "acknowledged")


Message.xep0333_marker = aioxmpp.xso.Child([
    ReceivedMarker,
    DisplayedMarker,
    AcknowledgedMarker,
])

Message.xep0333_markable = aioxmpp.xso.ChildFlag(
    (namespaces.xep0333_markers, "markable"),
)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.adhoc` --- Ad-Hoc Commands support (:xep:`50`)
#############################################################

This subpackage implements support for Ad-Hoc Commands as specified in
:xep:`50`. Both the client and the server side of Ad-Hoc Commands are
supported.

.. versionadded:: 0.8

Client-side
===========

.. currentmodule:: aioxmpp

.. autoclass:: AdHocClient

.. currentmodule:: aioxmpp.adhoc.service

.. autoclass:: ClientSession

Server-side
===========

.. currentmodule:: aioxmpp.adhoc

.. autoclass:: AdHocServer

.. currentmodule:: aioxmpp.adhoc.service

..
    .. autoclass:: ServerSession

XSOs
====

.. currentmodule:: aioxmpp.adhoc.xso

.. autoclass:: Command

.. autoclass:: Actions

.. autoclass:: Note

.. currentmodule:: aioxmpp.adhoc

Enumerations
------------

.. autoclass:: CommandStatus

.. autoclass:: ActionType
"""

from .service import (  # NOQA
    AdHocClient,
    ClientSession,
    AdHocServer,
)

from .xso import (  # NOQA
    CommandStatus,
    ActionType,
)

from . import xso  # NOQA



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.adhoc

from aioxmpp.utils import namespaces

from aioxmpp.e2etest import (
    require_feature,
    blocking,
    blocking_timed,
    TestCase,
    skip_with_quirk,
    Quirk,
)


class TestAdHocClient(TestCase):
    @require_feature(namespaces.xep0050_commands, multiple=True)
    @blocking
    @asyncio.coroutine
    def setUp(self, commands_providers):
        services = [aioxmpp.AdHocClient]

        self.peers = commands_providers

        self.client, = yield from asyncio.gather(
            self.provisioner.get_connected_client(
                services=services,
            ),
        )
        self.svc = self.client.summon(aioxmpp.AdHocClient)

    @asyncio.coroutine
    def _get_ping_peer(self):
        for peer in self.peers:
            for item in (yield from self.svc.get_commands(peer)):
                if item.node == "ping":
                    return peer
        self.assertTrue(
            False,
            "found no peer which offers ad-hoc-ping; consider setting "
            "the #no-adhoc-ping quirk"
        )

    @blocking_timed
    @asyncio.coroutine
    def test_get_list(self):
        items = []
        for peer in self.peers:
            items.extend((yield from self.svc.get_commands(
                peer
            )))

        self.assertTrue(items)

    @skip_with_quirk(Quirk.NO_ADHOC_PING)
    @blocking_timed
    @asyncio.coroutine
    def test_ping(self):
        ping_peer = yield from self._get_ping_peer()
        session = yield from self.svc.execute(ping_peer, "ping")
        self.assertTrue(session.response.notes)
        yield from session.close()

    @skip_with_quirk(Quirk.NO_ADHOC_PING)
    @blocking_timed
    @asyncio.coroutine
    def test_ping_with_async_cm(self):
        # TODO: port this to python 3.5+ once we require its
        ping_peer = yield from self._get_ping_peer()
        session = yield from self.svc.execute(ping_peer, "ping")
        yield from session.__aenter__()
        self.assertTrue(session.response.notes)
        yield from session.__aexit__(None, None, None)


class TestAdHocServer(TestCase):
    @blocking
    @asyncio.coroutine
    def setUp(self):
        self.client, self.server = yield from asyncio.gather(
            self.provisioner.get_connected_client(
                services=[aioxmpp.adhoc.AdHocClient],
            ),
            self.provisioner.get_connected_client(
                services=[aioxmpp.adhoc.AdHocServer],
            ),
        )

        self.client_svc = self.server.summon(aioxmpp.adhoc.AdHocClient)
        self.server_svc = self.server.summon(aioxmpp.adhoc.AdHocServer)

    @asyncio.coroutine
    def _trivial_handler(self, stanza):
        return aioxmpp.adhoc.xso.Command(
            "simple",
            notes=[
                aioxmpp.adhoc.xso.Note(
                    aioxmpp.adhoc.xso.NoteType.INFO,
                    "some info!"
                )
            ]
        )

    @blocking_timed
    @asyncio.coroutine
    def test_advertises_command_support(self):
        self.assertTrue(
            (yield from self.client_svc.supports_commands(
                self.server.local_jid,
            ))
        )

    @blocking_timed
    @asyncio.coroutine
    def test_respond_to_command_listing(self):
        self.server_svc.register_stateless_command(
            "simple",
            {
                aioxmpp.structs.LanguageTag.fromstr("en"): "Simple command",
                aioxmpp.structs.LanguageTag.fromstr("de"): "Einfacher Befehl",
            },
            self._trivial_handler,
        )

        commands = yield from self.client_svc.get_commands(
            self.server.local_jid
        )
        self.assertEqual(len(commands), 1)

        self.assertEqual(
            commands[0].node,
            "simple",
        )

        self.assertEqual(
            commands[0].name,
            "Simple command",
        )

    @blocking_timed
    @asyncio.coroutine
    def test_respond_to_command_info_query(self):
        self.server_svc.register_stateless_command(
            "simple",
            {
                aioxmpp.structs.LanguageTag.fromstr("en"): "Simple command",
                aioxmpp.structs.LanguageTag.fromstr("de"): "Einfacher Befehl",
            },
            self._trivial_handler,
            features={"foo"},
        )

        info = yield from self.client_svc.get_command_info(
            self.server.local_jid,
            "simple",
        )

        self.assertSetEqual(
            info.features,
            {
                namespaces.xep0050_commands,
                namespaces.xep0030_info,
                "foo",
            }
        )

        self.assertCountEqual(
            [
                ("automation", "command-node",
                 aioxmpp.structs.LanguageTag.fromstr("en"), "Simple command"),
                ("automation", "command-node",
                 aioxmpp.structs.LanguageTag.fromstr("de"), "Einfacher Befehl"),
            ],
            [
                (ident.category, ident.type_, ident.lang, ident.name)
                for ident in info.identities
            ]
        )

    @blocking_timed
    @asyncio.coroutine
    def test_execute_simple_command(self):
        self.server_svc.register_stateless_command(
            "simple",
            {
                aioxmpp.structs.LanguageTag.fromstr("en"): "Simple command",
                aioxmpp.structs.LanguageTag.fromstr("de"): "Einfacher Befehl",
            },
            self._trivial_handler,
        )

        session = yield from self.client_svc.execute(
            self.server.local_jid,
            "simple",
        )

        self.assertEqual(
            len(session.response.notes),
            1
        )
        self.assertEqual(session.response.notes[0].body, "some info!")
        self.assertIsNone(session.first_payload)

        yield from session.close()

    @blocking_timed
    @asyncio.coroutine
    def test_properly_fail_for_unknown_command(self):
        with self.assertRaises(aioxmpp.XMPPCancelError) as ctx:
            session = yield from self.client_svc.execute(
                self.server.local_jid,
                "simple",
            )

        self.assertEqual(
            ctx.exception.condition,
            aioxmpp.ErrorCondition.ITEM_NOT_FOUND
        )



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.service

from aioxmpp.utils import namespaces

from . import xso as carbons_xso


class CarbonsClient(aioxmpp.service.Service):
    """
    Provide an interface to enable and disable Message Carbons on the server
    side.

    .. note::

       This service deliberately does not provide a way to actually obtain sent
       or received carbonated messages.

       The common way for a service to do this would be a stanza filter (see
       :class:`aioxmpp.stream.StanzaStream`); however, in general the use and
       further distribution of carbonated messages highly depends on the
       application: it does, for example, not make sense to simply unwrap
       carbonated messages.

    .. automethod:: enable

    .. automethod:: disable
    """

    ORDER_AFTER = [
        aioxmpp.DiscoClient,
    ]

    @asyncio.coroutine
    def _check_for_feature(self):
        disco_client = self.dependencies[aioxmpp.DiscoClient]
        info = yield from disco_client.query_info(
            self.client.local_jid.replace(
                localpart=None,
                resource=None,
            )
        )

        if namespaces.xep0280_carbons_2 not in info.features:
            raise RuntimeError(
                "Message Carbons ({}) are not supported by the server".format(
                    namespaces.xep0280_carbons_2
                )
            )

    @asyncio.coroutine
    def enable(self):
        """
        Enable message carbons.

        :raises RuntimeError: if the server does not support message carbons.
        :raises aioxmpp.XMPPError: if the server responded with an error to the
                                   request.
        :raises: as specified in :meth:`aioxmpp.Client.send`
        """
        yield from self._check_for_feature()

        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=carbons_xso.Enable()
        )

        yield from self.client.send(iq)

    @asyncio.coroutine
    def disable(self):
        """
        Disable message carbons.

        :raises RuntimeError: if the server does not support message carbons.
        :raises aioxmpp.XMPPError: if the server responded with an error to the
                                   request.
        :raises: as specified in :meth:`aioxmpp.Client.send`
        """
        yield from self._check_for_feature()

        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=carbons_xso.Disable()
        )

        yield from self.client.send(iq)



########################################################################
# File name: execute.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import html

import PyQt5.Qt as Qt

import aioxmpp.adhoc

from .utils import asyncify, asyncify_blocking
from .ui.form import Ui_FormDialog


class Executor(Qt.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_FormDialog()
        self.ui.setupUi(self)
        self.base_title = self.windowTitle()
        self.session = None

        self.action_buttons = {
            aioxmpp.adhoc.ActionType.NEXT: self.ui.btn_next,
            aioxmpp.adhoc.ActionType.PREV: self.ui.btn_prev,
            aioxmpp.adhoc.ActionType.CANCEL: self.ui.btn_cancel,
            aioxmpp.adhoc.ActionType.COMPLETE: self.ui.btn_complete,
        }

        self.ui.btn_close.clicked.connect(self._close)
        self.ui.btn_next.clicked.connect(self._next)
        self.ui.btn_prev.clicked.connect(self._prev)
        self.ui.btn_cancel.clicked.connect(self._cancel)
        self.ui.btn_complete.clicked.connect(self._complete)

        self.fieldmap = {}
        self.form_area = None

    def _close(self):
        self.close()

    def _next(self):
        self._action(aioxmpp.adhoc.ActionType.NEXT)

    def _prev(self):
        self._action(aioxmpp.adhoc.ActionType.PREV)

    def _cancel(self):
        self.close()

    def _complete(self):
        self._action(aioxmpp.adhoc.ActionType.COMPLETE)

    @asyncify_blocking
    @asyncio.coroutine
    def run_with_session_fut(self, name, session_fut):
        self.setWindowTitle("{} - {}".format(
            name,
            self.base_title
        ))
        self.show()
        try:
            self.session = yield from session_fut
        except Exception as exc:
            self.fail(exc)
            return
        self.response_received()

    def _update_buttons(self):
        status = self.session.status
        if status != aioxmpp.adhoc.CommandStatus.EXECUTING:
            self.ui.btn_cancel.hide()
            self.ui.btn_next.hide()
            self.ui.btn_prev.hide()
            self.ui.btn_complete.hide()
            self.ui.btn_close.show()
        else:
            self.ui.btn_close.hide()
            allowed_actions = self.session.allowed_actions
            for action, btn in self.action_buttons.items():
                if action in allowed_actions:
                    btn.show()
                else:
                    btn.hide()

    def _update_notes(self):
        notes = self.session.response.notes
        if not notes:
            self.ui.notes_area.hide()
            return

        self.ui.notes_area.show()
        source_parts = []
        for note in notes:
            print(note.type_, note.body)
            source_parts.append("<p><b>{}: </b>{}</p>".format(
                html.escape(note.type_.value),
                "</p><p>".join(html.escape(note.body or "").split("\n"))
            ))

        self.ui.notes_area.setText("\n".join(source_parts))

    def _update_form(self):
        payload = self.session.first_payload
        if not isinstance(payload, aioxmpp.forms.Data):
            self.ui.form_widget.hide()
            return

        self.ui.form_widget.show()
        if payload.title:
            self.ui.title.show()
            self.ui.title.setText(payload.title)
        else:
            self.ui.title.hide()

        if payload.instructions:
            self.ui.instructions.show()
            self.ui.instructions.setText(
                "<p>{}</p>".format(
                    "</p><p>".join(map(html.escape, payload.instructions))
                )
            )
        else:
            self.ui.instructions.hide()

        if self.form_area is not None:
            self.ui.form_widget.layout().removeWidget(self.form_area)
            self.ui.form_widget.children().remove(self.form_area)
            self.form_area.deleteLater()

        self.form_area = Qt.QScrollArea()
        layout = Qt.QFormLayout()
        self.form_area.setLayout(layout)
        self.ui.form_widget.layout().addWidget(self.form_area)

        self.fieldmap = {}
        for field in payload.fields:
            if field.var == "FORM_TYPE":
                continue
            label, widget = None, None
            if field.type_ == aioxmpp.forms.FieldType.FIXED:
                label = Qt.QLabel(field.values[0])
                layout.addRow(
                    label
                )
            elif field.type_ in {aioxmpp.forms.FieldType.LIST_SINGLE,
                                 aioxmpp.forms.FieldType.LIST_MULTI}:
                label = Qt.QLabel(field.label)
                widget = Qt.QListWidget()
                for opt_value, opt_label in sorted(field.options.items()):
                    item = Qt.QListWidgetItem(opt_label, widget)
                    item.setData(Qt.Qt.UserRole, opt_value)
                    widget.addItem(item)

                if field.type_.is_multivalued:
                    widget.setSelectionMode(
                        Qt.QAbstractItemView.MultiSelection
                    )
                else:
                    widget.setSelectionMode(
                        Qt.QAbstractItemView.SingleSelection
                    )
                layout.addRow(label, widget)

            elif field.type_ in {aioxmpp.forms.FieldType.TEXT_SINGLE,
                                 aioxmpp.forms.FieldType.JID_SINGLE}:
                label = Qt.QLabel(field.label)
                widget = Qt.QLineEdit()
                if field.values:
                    widget.setText(field.values[0])
                layout.addRow(label, widget)

            elif field.type_ in {aioxmpp.forms.FieldType.TEXT_PRIVATE}:
                label = Qt.QLabel(field.label)
                widget = Qt.QLineEdit()
                widget.setEchoMode(Qt.QLineEdit.Password)
                widget.setInputMethodHints(Qt.Qt.ImhHiddenText |
                                           Qt.Qt.ImhNoPredictiveText |
                                           Qt.Qt.ImhNoAutoUppercase)
                if field.values:
                    widget.setText(field.values[0])
                layout.addRow(label, widget)

            elif field.type_ in {aioxmpp.forms.FieldType.TEXT_MULTI,
                                 aioxmpp.forms.FieldType.JID_MULTI}:
                label = Qt.QLabel(field.label)
                widget = Qt.QTextEdit()
                widget.setText("\n".join(field.values))
                widget.setAcceptRichText(False)
                layout.addRow(label, widget)

            else:
                self.fail("unhandled field type: {}".format(field.type_))

            self.fieldmap[field.var] = label, widget

    def _fill_form(self):
        for field in self.session.first_payload.fields:
            try:
                _, widget = self.fieldmap[field.var]
            except KeyError:
                continue
            if widget is None:
                continue

            if field.type_ in {aioxmpp.forms.FieldType.LIST_SINGLE,
                               aioxmpp.forms.FieldType.LIST_MULTI}:
                # widget is list widget
                selected = []
                for index in widget.selectionModel().selectedIndexes():
                    selected.append(
                        widget.model().data(
                            index,
                            Qt.Qt.UserRole
                        )
                    )

                field.values[:] = selected

            elif field.type_ in {aioxmpp.forms.FieldType.JID_SINGLE,
                                 aioxmpp.forms.FieldType.TEXT_SINGLE,
                                 aioxmpp.forms.FieldType.TEXT_PRIVATE}:
                # widget is line edit
                field.values[:] = [widget.text()]

            elif field.type_ in {aioxmpp.forms.FieldType.JID_MULTI,
                                 aioxmpp.forms.FieldType.TEXT_MULTI}:
                field.values[:] = widget.toPlainText().split("\n")

    def _action(self, type_):
        self._fill_form()
        self._submit_action(type_)

    @asyncify_blocking
    @asyncio.coroutine
    def _submit_action(self, type_):
        yield from self.session.proceed(action=type_)
        self.response_received()

    def response_received(self):
        try:
            self.ui.status.setText(str(self.session.status))
            self._update_buttons()
            self._update_notes()
            self._update_form()
        except Exception as exc:
            self.fail(str(exc))

    def fail(self, message):
        Qt.QMessageBox.critical(
            self.parent(),
            "Error",
            message,
        )
        self.close()

    @asyncify
    def closeEvent(self, ev):
        if self.session is not None:
            yield from self.session.close()
        return super().closeEvent(ev)
